module Matriz where

-- =========================================================================
-- Esse arquivo define a matriz e as funcoes referentes a ela que serao utilizadas
-- no projeto
-- =========================================================================

-- ==============================================================================
-- Matriz eh uma abstracao da Array Bidimencional, com metodos que nos auxiliarao
-- a trabalhar com ela
data Matriz a = Matriz [[a]] deriving Eq

-- ==============================================================================
-- A instancia de show foi definida manualmente para facilitar o debug
instance (Show a) => Show (Matriz a) where
    show (Matriz (linha:[]))     = show linha
    show (Matriz (linha:linhas)) = (show linha) ++ "\n" ++ show (Matriz linhas)

-- ==============================================================================
-- A instancia de semigroup existe pois o <> eh usado para concatenar duas matrizes
-- em algumas funcoes
instance Semigroup (Matriz a) where
    (Matriz m1) <> (Matriz m2) = Matriz (m1++m2)

-- ==============================================================================
-- Retorna o valor da matriz no indice (i, j)
valorEm :: Matriz a -> Int -> Int -> a
valorEm (Matriz m) i j = m !! i !! j

-- ==============================================================================
-- Retorna a matriz armazenada na estrutura
getMatriz :: Matriz a -> [[a]]
getMatriz (Matriz m) = m

-- ==============================================================================
-- Coloca o valor val no indice (i, j) da matriz
colocaEm ::  a -> Int -> Int -> Matriz a -> Matriz a
colocaEm val i j (Matriz (l:ls))
    | i < 0     = (Matriz (l:ls))
    | i == 0    = Matriz (l':ls)
    | otherwise = (Matriz [l]) <> (colocaEm val (i-1) j (Matriz ls))
        where
            l' = colocaNaLinha l val j

colocaNaLinha :: [a] -> a -> Int -> [a]
colocaNaLinha (_:xs) val 0 = val:xs
colocaNaLinha (x:xs) val n = x:(colocaNaLinha xs val (n-1))

-- ==============================================================================
-- Retorna as dimensoes da matriz
alturaMatriz :: Matriz a -> Int
alturaMatriz (Matriz m) = length m

larguraMatriz :: Matriz a -> Int
larguraMatriz (Matriz m) = length $ m !! 0

-- ==============================================================================
-- Rotaciona uma matriz no sentido horario uma vez
rotacionaDir :: Matriz a -> Matriz a
rotacionaDir (Matriz m) = Matriz [(col2LinhaDir m (i-1)) | i <-[1..(length $ m !! 0)]]

-- Funcao auxiliar que transforma uma coluna da matriz original na linha da matriz
-- rotacionada. Note que na rotacao a direita eh necessario pegar os valores de baixo
-- para cima, por isso o reverse
col2LinhaDir :: [[a]] -> Int -> [a]
col2LinhaDir m i = reverse [(m !! (j-1) !! i) | j <- [1..(length m)]]

-- ==============================================================================
-- Rotaciona uma matriz no sentido anti-horario uma vez
rotacionaEsq :: Matriz a -> Matriz a
rotacionaEsq (Matriz m) = Matriz [col2LInhaEsq m (i-1) | i <- (reverse [1..(length $ m !! 0)])]

-- Funcao auxiliar que transforma uma coluna da matriz original na linha da matriz
-- rotacionada
col2LInhaEsq :: [[a]] -> Int -> [a]
col2LInhaEsq m i = [(m !! (j-1) !! i) | j <- [1..(length m)]]